import numpy as np
import onnxruntime as ort
import cv2
import os
import argparse
from glob import glob
from typing import List, Dict, Optional
import time
from datetime import datetime
from tqdm import tqdm
from src.analysis.report_generator import ReportGenerator
import onnx

class ONNXDetector:
    def select_input_directory(self) -> Optional[str]:
        """图形界面选择输入目录"""
        import tkinter as tk
        from tkinter import filedialog
        root = tk.Tk()
        root.withdraw()
        folder_path = filedialog.askdirectory(title='选择输入目录')
        return folder_path if folder_path else None

    def __init__(self, threshold: float = 0.5, 
                 output_dir: str = None, save_empty: bool = False, 
                 max_bbox_ratio: float = 0.5,
                 input_dir: str = None, model_type: str = None, model_path: str = None):
        self.image_count = 0
        self.detection_records = []
        self.input_dir = input_dir
        self.save_empty = save_empty
        self.threshold = threshold
        self.confThreshold = threshold
        
        # 自动识别模型类型
        if model_path:
            self.model_path = model_path
            self.model_type = model_type if model_type else get_model_type(model_path)
        else:
            self.model_type = model_type if model_type else get_model_type(r"D:\PythonProject\R360-UAVmodelTool\src\models\UAV-250411.onnx")
            # 根据模型类型选择不同的模型路径
            if self.model_type == 'Anti_UAV':
                self.model_path = r"D:\PythonProject\R360-UAVmodelTool\src\models\250411_Anti_UAV.onnx"
            elif self.model_type == 'UAV-250411':
                self.model_path = r"D:\PythonProject\R360-UAVmodelTool\src\models\UAV-250411.onnx"
            else:
                self.model_path = r"D:\PythonProject\R360-UAVmodelTool\src\models\uav_and_bird.onnx"
        # 模型初始化完成
        
        # 初始化ONNX会话
        so = ort.SessionOptions()
        so.log_severity_level = 3
        self.providers = ['CUDAExecutionProvider', 'CPUExecutionProvider']
        self.session = ort.InferenceSession(self.model_path, so, providers=self.providers)
        
        # 模型参数设置
        self.input_size = (640, 640)  # width, height
        self.mean_ = np.array([0.485, 0.456, 0.406], dtype=np.float32)
        self.std_ = np.array([0.229, 0.224, 0.225], dtype=np.float32)
        self.max_bbox_ratio = max_bbox_ratio
        
        # 从模型路径解析版本号
        self.model_version = os.path.basename(self.model_path).split('.')[0]
        
        # 获取输入输出名称
        self.input_name = self.session.get_inputs()[0].name
        if self.model_type == 'Anti_UAV':
            self.output_name = 'multiclass_nms3_0.tmp_0'
        elif self.model_type == 'UAV-250411':
            self.output_name = 'tmp_17'
        else:
            self.output_name = 'conv2d_308.tmp_1'
        
        # 类别映射
        if self.model_type == 'uav_and_bird':
            self.class_names = {0: 'Bird', 1: 'UAV'}
        else:
            self.class_names = {0: 'UAV'}

        # 误报特征库相关代码已移除
        
        # 创建输出目录
        output_base = os.path.join(os.path.dirname(__file__), '..', '..', 'Output')
        os.makedirs(output_base, exist_ok=True)
        
        timestamp = datetime.now().strftime('%m%d%H%M%S')
        if output_dir is None and input_dir is not None:
            self.output_dir = os.path.join(output_base, f"{os.path.basename(os.path.normpath(input_dir))}_{timestamp}")
        else:
            self.output_dir = os.path.join(output_base, output_dir or f"output_{timestamp}")
            
        # 创建目标存储目录
        self.targets_dir = os.path.join(self.output_dir, 'targets')
        os.makedirs(self.targets_dir, exist_ok=True)

    def preprocess(self, image_path: str) -> Optional[np.ndarray]:
        """图像预处理"""
        # 初始化CUDA加速
        if not hasattr(self, 'use_cuda'):
            self.use_cuda = cv2.cuda.getCudaEnabledDeviceCount() > 0
            self.gpu_frame = cv2.cuda_GpuMat() if self.use_cuda else None

        # 读取图像到GPU（如果可用）
        if self.use_cuda:
            self.gpu_frame.upload(cv2.imread(image_path))
            if self.gpu_frame.empty():
                # 图像加载失败
                return None
            
            # GPU预处理流水线
            gpu_resized = cv2.cuda.resize(self.gpu_frame, self.input_size)
            gpu_rgb = cv2.cuda.cvtColor(gpu_resized, cv2.COLOR_BGR2RGB)
            
            # 下载到CPU进行后续处理
            image_orig = gpu_rgb.download()
            self.orig_h, self.orig_w = image_orig.shape[:2]
        else:
            # CPU回退路径
            image_orig = cv2.imread(image_path)
            if image_orig is None:
                # 图像加载失败
                return None
            
            self.orig_h, self.orig_w = image_orig.shape[:2]
            image_orig = cv2.resize(image_orig, self.input_size)
            image_orig = cv2.cvtColor(image_orig, cv2.COLOR_BGR2RGB)

        # 统一的后处理
        image = image_orig.astype(np.float32) / 255.0
        image -= self.mean_[None, None, :]
        image /= self.std_[None, None, :]
        image = np.transpose(image, (2, 0, 1))  # CHW 格式
        return np.expand_dims(image, axis=0)

    def nms(self, boxes: np.ndarray, scores: np.ndarray, conf_threshold: float, iou_threshold: float) -> List[int]:
        """非极大值抑制"""
        x1 = boxes[:, 0]
        y1 = boxes[:, 1]
        x2 = boxes[:, 2]
        y2 = boxes[:, 3]
        areas = (x2 - x1 + 1) * (y2 - y1 + 1)
        order = scores.argsort()[0][::-1]
        keep = []
        
        while order.size > 0:
            i = order[0]
            keep.append(i)
            xx1 = np.maximum(x1[i], x1[order[1:]])
            yy1 = np.maximum(y1[i], y1[order[1:]])
            xx2 = np.minimum(x2[i], x2[order[1:]])
            yy2 = np.minimum(y2[i], y2[order[1:]])
            w = np.maximum(0.0, xx2 - xx1 + 1)
            h = np.maximum(0.0, yy2 - yy1 + 1)
            inter = w * h
            ovr = inter / (areas[i] + areas[order[1:]] - inter)
            inds = np.where(ovr <= iou_threshold)[0]
            order = order[inds + 1]
        return keep

    def inference(self, input_data: np.ndarray):
        if self.model_type == 'uav_and_bird':
            # 返回所有输出
            outputs = self.session.run(None, {self.input_name: input_data})
            # 模型推理完成
            return outputs
        elif self.model_type in ['Anti_UAV', 'UAV-250411']:
            scale_factor = np.array([[1, 1]], dtype=np.float32)
            output = self.session.run(
                [self.output_name],
                {self.input_name: input_data, 'scale_factor': scale_factor}
            )[0]
            # 模型推理完成
            return output
        else:
            output = self.session.run(
                [self.output_name],
                {self.input_name: input_data}
            )[0]
            # 模型推理完成
            return output

    # ORB特征匹配误报检测方法已移除

    def postprocess(self, detections, image_orig: np.ndarray, image_path: str) -> tuple:
        # 开始后处理
        
        valid_detections = 0
        img_out = image_orig.copy()
        detections_list = []
        
        if self.model_type == 'Anti_UAV':
            # Anti_UAV模型的后处理逻辑
            keep_idx = (detections[:, 1] > self.confThreshold)
            detections = detections[keep_idx]
            
            if len(detections) == 0:
                return 0, img_out, []
            ratioh = self.orig_h / self.input_size[1]
            ratiow = self.orig_w / self.input_size[0]
            
            detections[:, 2:6] *= np.array([ratiow, ratioh, ratiow, ratioh])
            keep = self.nms(detections[:, 2:6], detections[:, 1:2], self.confThreshold, 0.4)
            
            for idx in keep:
                class_id = int(detections[idx, 0])
                confidence = detections[idx, 1]
                x1, y1, x2, y2 = detections[idx, 2:6].astype(int)
                bbox_area = (x2 - x1) * (y2 - y1)
                image_area = self.orig_w * self.orig_h
                
                if bbox_area / image_area > self.max_bbox_ratio:
                    continue
                roi = image_orig[y1:y2, x1:x2]
                if roi.size == 0:
                    continue
                # 误报过滤已移除
                target_filename = f"{os.path.splitext(os.path.basename(image_path))[0]}_{valid_detections}.jpg"
                cv2.imwrite(os.path.join(self.targets_dir, target_filename), roi)
                label = f'{self.class_names[class_id]} {confidence:.2f}'
                cv2.rectangle(img_out, (x1, y1), (x2, y2), (255, 0, 0), 4)
                cv2.putText(img_out, label, (x1, y1 - 10), 
                            cv2.FONT_HERSHEY_SIMPLEX, 0.8, (0, 0, 255), 2)
                valid_detections += 1
                detections_list.append({
                    'class_id': class_id,
                    'confidence': float(confidence),
                    'bbox': [x1, y1, x2, y2],
                    'orig_w': self.orig_w,
                    'orig_h': self.orig_h
                })
        elif self.model_type == 'UAV-250411':
            # UAV-250411模型的后处理逻辑
            keep_idx = (detections[:, 1] > self.confThreshold)
            detections = detections[keep_idx]
            
            if len(detections) == 0:
                return 0, img_out, []
            ratioh = self.orig_h / self.input_size[1]
            ratiow = self.orig_w / self.input_size[0]
            
            detections[:, 2:6] *= np.array([ratiow, ratioh, ratiow, ratioh])
            keep = self.nms(detections[:, 2:6], detections[:, 1:2], self.confThreshold, 0.4)
            
            for idx in keep:
                class_id = int(detections[idx, 0])
                confidence = detections[idx, 1]
                x1, y1, x2, y2 = detections[idx, 2:6].astype(int)
                bbox_area = (x2 - x1) * (y2 - y1)
                image_area = self.orig_w * self.orig_h
                
                if bbox_area / image_area > self.max_bbox_ratio:
                    continue
                roi = image_orig[y1:y2, x1:x2]
                if roi.size == 0:
                    continue
                # 误报过滤已移除
                target_filename = f"{os.path.splitext(os.path.basename(image_path))[0]}_{valid_detections}.jpg"
                cv2.imwrite(os.path.join(self.targets_dir, target_filename), roi)
                label = f'{self.class_names[class_id]} {confidence:.2f}'
                cv2.rectangle(img_out, (x1, y1), (x2, y2), (255, 0, 0), 4)
                cv2.putText(img_out, label, (x1, y1 - 10), 
                            cv2.FONT_HERSHEY_SIMPLEX, 0.8, (0, 0, 255), 2)
                valid_detections += 1
                detections_list.append({
                    'class_id': class_id,
                    'confidence': float(confidence),
                    'bbox': [x1, y1, x2, y2],
                    'orig_w': self.orig_w,
                    'orig_h': self.orig_h
                })
        elif self.model_type == 'uav_and_bird':
            # 根据用户提供的详细说明，模型输出9个张量，按3个尺度分组：
            # 小尺度 (20x20): 输出0(bbox), 输出1(cls), 输出2(quality)
            # 中等尺度 (40x40): 输出3(bbox), 输出4(cls), 输出5(quality) 
            # 大尺度 (80x80): 输出6(bbox), 输出7(cls), 输出8(quality)
            if len(detections) != 9:
                return 0, img_out, []
            
            # 定义尺度信息：(bbox_idx, cls_idx, quality_idx, grid_size, stride)
            scales = [
                (0, 1, 2, 20, 32),  # 小尺度特征图，用于检测大物体
                (3, 4, 5, 40, 16),  # 中等尺度特征图，用于检测中等物体
                (6, 7, 8, 80, 8),   # 大尺度特征图，用于检测小物体
            ]
            
            all_boxes, all_scores, all_classes = [], [], []
            
            for scale_idx, (bbox_idx, cls_idx, quality_idx, grid_size, stride) in enumerate(scales):
                # 获取当前尺度的输出
                bbox_output = detections[bbox_idx]      # (1, 68, H, W)
                cls_output = detections[cls_idx]        # (1, 2, H, W) 
                quality_output = detections[quality_idx] # (1, 1, H, W)
                
                # 移除batch维度
                bbox_pred = bbox_output[0]    # (68, H, W)
                cls_pred = cls_output[0]      # (2, H, W)
                quality_pred = quality_output[0] # (1, H, W)
                
                H, W = bbox_pred.shape[1], bbox_pred.shape[2]
                
                # 创建网格坐标 - 添加0.5偏移表示像素中心
                grid_y, grid_x = np.meshgrid(np.arange(H), np.arange(W), indexing='ij')
                grid_x = (grid_x + 0.5).flatten()  # (H*W,) - 像素中心偏移
                grid_y = (grid_y + 0.5).flatten()  # (H*W,) - 像素中心偏移
                
                # 重塑张量为 (H*W, channels)
                bbox_pred = bbox_pred.reshape(68, -1).T    # (H*W, 68)
                cls_pred = cls_pred.reshape(2, -1).T       # (H*W, 2)
                quality_pred = quality_pred.reshape(-1)    # (H*W,)
                
                # 解码边界框 - Distribution Focal Loss解码
                # 68个通道 = 4个边界 * 17个分布值
                bbox_pred = bbox_pred.reshape(-1, 4, 17)  # (H*W, 4, 17)
                
                # 对每个边界的17个值应用softmax
                bbox_pred_softmax = np.exp(bbox_pred) / np.sum(np.exp(bbox_pred), axis=2, keepdims=True)
                
                # 创建范围向量 [0, 1, 2, ..., 16]
                range_vector = np.arange(17).reshape(1, 1, 17)
                
                # 计算期望距离值
                distances = np.sum(bbox_pred_softmax * range_vector, axis=2)  # (H*W, 4)
                d_left, d_top, d_right, d_bottom = distances[:, 0], distances[:, 1], distances[:, 2], distances[:, 3]
                
                # 计算最终边界框坐标
                x1 = (grid_x - d_left) * stride
                y1 = (grid_y - d_top) * stride
                x2 = (grid_x + d_right) * stride
                y2 = (grid_y + d_bottom) * stride
                
                # 组合边界框
                boxes = np.stack([x1, y1, x2, y2], axis=1)  # (H*W, 4)
                
                # 计算最终置信度分数
                # cls_pred已经是sigmoid激活后的结果，quality_pred是clip后的结果
                final_scores = cls_pred * quality_pred[:, None]  # (H*W, 2)
                
                # 获取最大置信度和对应类别
                max_scores = np.max(final_scores, axis=1)  # (H*W,)
                class_ids = np.argmax(final_scores, axis=1)  # (H*W,)
                
                # 置信度过滤
                mask = max_scores > self.confThreshold
                
                if np.sum(mask) > 0:
                    filtered_boxes = boxes[mask]
                    filtered_scores = max_scores[mask]
                    filtered_classes = class_ids[mask]
                    
                    all_boxes.append(filtered_boxes)
                    all_scores.append(filtered_scores)
                    all_classes.append(filtered_classes)
            
            # 检查是否有有效检测
            if len(all_boxes) == 0 or all([len(b) == 0 for b in all_boxes]):
                return 0, img_out, []
            
            # 合并所有尺度的检测结果
            final_boxes = np.concatenate(all_boxes, axis=0)
            final_scores = np.concatenate(all_scores, axis=0)
            final_classes = np.concatenate(all_classes, axis=0)
            
            # 坐标缩放到原图尺寸
            ratioh = self.orig_h / self.input_size[1]
            ratiow = self.orig_w / self.input_size[0]
            
            final_boxes[:, [0, 2]] *= ratiow  # x坐标
            final_boxes[:, [1, 3]] *= ratioh  # y坐标
            
            # NMS处理
            keep = self.nms(final_boxes, final_scores[:, None], self.confThreshold, 0.4)
            
            # 处理最终检测结果
            for idx in keep:
                x1, y1, x2, y2 = final_boxes[idx].astype(int)
                confidence = final_scores[idx]
                class_id = int(final_classes[idx])
                
                # 确保坐标在图像范围内
                x1 = max(0, min(x1, self.orig_w - 1))
                y1 = max(0, min(y1, self.orig_h - 1))
                x2 = max(0, min(x2, self.orig_w - 1))
                y2 = max(0, min(y2, self.orig_h - 1))
                
                bbox_area = (x2 - x1) * (y2 - y1)
                image_area = self.orig_w * self.orig_h
                
                if bbox_area / image_area > self.max_bbox_ratio:
                    continue
                    
                if x2 <= x1 or y2 <= y1:
                    continue
                    
                roi = image_orig[y1:y2, x1:x2]
                if roi.size == 0:
                    continue
                    
                # 误报过滤已移除
                target_filename = f"{os.path.splitext(os.path.basename(image_path))[0]}_{valid_detections}.jpg"
                cv2.imwrite(os.path.join(self.targets_dir, target_filename), roi)
                label = f'{self.class_names[class_id]} {confidence:.2f}'
                cv2.rectangle(img_out, (x1, y1), (x2, y2), (255, 0, 0), 4)
                cv2.putText(img_out, label, (x1, y1 - 10), 
                            cv2.FONT_HERSHEY_SIMPLEX, 0.8, (0, 0, 255), 2)
                valid_detections += 1
                detections_list.append({
                    'class_id': class_id,
                    'confidence': float(confidence),
                    'bbox': [x1, y1, x2, y2],
                    'orig_w': self.orig_w,
                    'orig_h': self.orig_h
                })
                
        return valid_detections, img_out, detections_list

    def process_image(self, image_path: str) -> int:
        """端到端处理单张图像"""
        input_data = self.preprocess(image_path)
        if input_data is None:
            return 0
            
        detections = self.inference(input_data)
        valid_count, processed_img, detections_list = self.postprocess(detections, cv2.imread(image_path), image_path)
        
        # 无论是否检测到目标都记录信息
        record = {
            'image_path': image_path,
            'has_detection': valid_count > 0,
            'timestamp': datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        }

        filename = f"{os.path.splitext(os.path.basename(image_path))[0]}.jpg"
        # 创建输出图片目录
        output_img_dir = os.path.join(self.output_dir, 'imgs')
        os.makedirs(output_img_dir, exist_ok=True)
        output_path = os.path.join(output_img_dir, filename)
        
        # 根据设置保存图片
        if valid_count > 0 or self.save_empty:
            cv2.imwrite(output_path, processed_img)
            # 输出图像已保存
        else:
            output_path = None
            # 未保存图像 (无检测且save_empty=False)
            
        self.image_count += 1
        
        # 记录检测信息
        if valid_count > 0:
            record['detections'] = detections_list
        
        self.detection_records.append(record)
        # 处理完成
        return valid_count

def get_model_type(model_path: str) -> str:
    """
    根据ONNX模型的输入输出结构自动判断模型类型。
    返回值：
        'Anti_UAV'：250411_Anti_UAV.onnx
        'UAV-250411'：UAV-250411.onnx
        'uav_and_bird'：uav_and_bird.onnx
        'unknown'：无法识别
    """
    try:
        model = onnx.load(model_path)
        input_names = [i.name for i in model.graph.input]
        output_names = [o.name for o in model.graph.output]
        # 判断Anti_UAV模型
        if 'scale_factor' in input_names:
            if any(name.startswith('multiclass_nms3') for name in output_names):
                return 'Anti_UAV'
            elif len(output_names) == 2 and output_names[0].startswith('tmp_'):
                return 'UAV-250411'
            else:
                return 'UAV-250411'
        else:
            # 没有scale_factor输入，且输出较多
            if len(output_names) > 2:
                return 'uav_and_bird'
        return 'unknown'
    except Exception as e:
        # 模型类型识别失败
        pass
        return 'unknown'

# 命令行接口
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='UAV Detection using ONNX model')
    parser.add_argument('--input', type=str, help='输入图像路径或目录')
    parser.add_argument('--threshold', type=float, default=0.5, help='检测置信度阈值')
    parser.add_argument('--output', type=str, default=None, help='输出目录路径，默认为输入目录名+_results')
    parser.add_argument('--max-bbox-ratio', type=float, default=0.05,
                        help='检测框最大面积比例阈值，默认0.05')
    parser.add_argument('--save-empty', action='store_true',
                        help='是否保存未检测到目标的图片')
    parser.add_argument('--gui', action='store_true',
                        help='启用图形界面选择输入目录')
    parser.add_argument('--model-type', type=str, choices=['Anti_UAV', 'UAV-250411', 'uav_and_bird'],
                        help='指定模型类型，不指定则自动识别')
    parser.add_argument('--model-path', type=str, help='指定模型路径，不指定则根据模型类型自动选择')
    args = parser.parse_args()

    # GUI模式处理
    if args.gui:
        detector = ONNXDetector(
            threshold=args.threshold,
            max_bbox_ratio=args.max_bbox_ratio,
            output_dir=args.output,
            save_empty=args.save_empty,
            model_type=args.model_type,
            model_path=args.model_path
        )
        input_dir = detector.select_input_directory()
        if not input_dir:
            # 未选择目录，程序退出
            exit()
        args.input = input_dir
    
    # 初始化检测器时传递参数
    detector = ONNXDetector(
        threshold=args.threshold,
        max_bbox_ratio=args.max_bbox_ratio,
        output_dir=args.output,
        save_empty=args.save_empty,
        input_dir=args.input if os.path.isdir(args.input) else None,
        model_type=args.model_type,
        model_path=args.model_path
    )

    def process_single(image_path: str):
        detections = detector.process_image(image_path)
        # 图像处理完成

    if os.path.isdir(args.input):
        total = 0
        image_files = []
        for root, _, files in os.walk(args.input):
            image_files.extend([os.path.join(root, f) for f in files if f.lower().endswith('.jpg')])
        for img_file in tqdm(image_files, desc='Processing images'):
            total += detector.process_image(img_file)
        # 批量处理完成
    
        # 生成CSV报告
        csv_path = os.path.join(detector.output_dir, f'detection_report_{detector.model_type}.csv')
        ReportGenerator(detector).generate_csv(csv_path)
        # CSV报告已生成
    else:
        detections = detector.process_image(args.input)